#include "headers.h"
#include "Rs232c.h"

#define LOGNAME	"rs232c"
#include "diag.h"

Rs232c::Rs232c()
: Device(DEV_ID('2', '3', '2', 'C')), access(0), scheduler(0),
  hcom(INVALID_HANDLE_VALUE), hthread(NULL), checking(false), setting(false)
{
	InitializeCriticalSection(&cs);
	config.baudrate = BAUDRATE_DEFAULT;
	config.comport  = COM_DEFAULT;
	clock = 16;
	parity = Parity::none;
	datalen = 8;
	stop = 1;
}

Rs232c::~Rs232c()
{
	DeleteCriticalSection(&cs);
}

bool Rs232c::Init(IIOAccess* _access, IScheduler* _scheduler, uint _prxrdy, uint _preq)
{
	access = _access;
	scheduler = _scheduler;
	prxrdy = _prxrdy;
	prequest =_preq;
	port30 = 0;
	Close();
	Reset();
	return true;
}

// ---------------------------------------------------------------------------
//	I/O |[gĎ

void IFCALL Rs232c::Reset(uint, uint)
{
	datacount = 0;
	datapos = 0;
	statuscount = 0;
	mode = Mode::clear;
	comstat.fDsrHold = 0;
	comstat.fTxim = 0;
}

void IFCALL Rs232c::Out20(uint, uint data)
{
	if (IsRs232cEnabled())
	{
		DWORD len;
		if (WriteFile(hcom, &data, 1, &len, 0))
		{
			//LOG2("\t\tW:%c %02X\n", (isprint(data) ? data : '.'), data);
			//char buf[256];
			//sprintf(buf,"W:%02X\r\n",data);
			//OutputDebugString(buf);
			statuscount = 0;
			DWORD error;
			ClearCommError(hcom, &error, &comstat);
			if (len == 0) 
			{
				LOG0("Write length 0\n");
			}
		}
		else 
		{
			LOG1("Write error %08X\n", GetLastError());
		}
	}
}

void IFCALL Rs232c::Out21(uint, uint d)
{
	//LOG1("[%.2x] ", d);
	
	switch (mode)
	{
	case clear:
		// Mode Instruction
		if (d & 3)
		{
			// Asynchronus mode
			mode = async;
			// b7 b6 b5 b4 b3 b2 b1 b0
			// STOP  EP PE CHAR  RATE
			static const int clockdiv[] = { 1, 1, 16, 64 };
			uint new_clock = clockdiv[d & 3];
			uint new_datalen = 5 + ((d >> 2) & 3);
			Parity new_parity = d & 0x10 ? (d & 0x20 ? even : odd) : none;
			uint new_stop = (d >> 6) & 3;
			if (clock != new_clock || datalen != new_datalen || 
				parity != new_parity || stop != new_stop)
			{
				clock = new_clock;
				datalen = new_datalen;
				parity = new_parity;
				stop = new_stop;
				SetMode(config.baudrate);
			}
			LOG4("Async: %d baud, Parity:%c Data:%d Stop:%s\n", clock, parity, datalen, stop==3 ? "2" : stop==2 ? "1.5" : "1");
		}
		else
		{
			// Synchronus mode
			mode = sync1;
			//clock = 0;
			Parity new_parity = d & 0x10 ? (d & 0x20 ? even : odd) : none;
			if (parity != new_parity || clock != 16)
			{
				parity = new_parity;
				clock = 16;
				SetMode(config.baudrate);
			}
			LOG2("Sync: %d baud, Parity:%c / ", clock, parity);
		}
		break;
		
	case sync1:
		mode = sync2;
		break;

	case sync2:
		mode = sync;
		LOG0("\n");
		break;

	case async:
	case sync:
		// Command Instruction
		// b7 - enter hunt mode
		// b6 - internal reset
		if (d & 0x40)
		{
			// Reset!
			LOG0(" Internal Reset!\n");
			Reset();
			mode = clear;
			access->In(0x20);	// clear RXRDY
			if (IsRs232cEnabled())
			{
				PurgeComm(hcom, 
					PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR);
			}
			break;
		}
		// b5 - request to send
		// b4 - error reset
		if (d & 0x10)
		{
			//LOG0(" ERRCLR"); 
			status &= ~(PE | OE | FE);
		}
		// b3 - send break charactor
		if (d & 8)
		{
			LOG0(" SNDBRK");
		}
		// b2 - receive enable 
		rxen = (d & 4) != 0;
		// b1 - data terminal ready
		// b0 - send enable
		txen = (d & 1) != 0;

		//LOG2(" RxE:%d TxE:%d\n", rxen, txen);
		break;
	default:
		LOG1("internal error? <%d>\n", mode);
		break;
	}
}

void IFCALL Rs232c::Out30(uint, uint data)
{
	// BIT 5,4	00...CMT 600baud
	//			01...CMT 1200baud
	//			10...RS232C ASYNC
	//			11...RS232C SYNC
	port30 = data;
}

void IFCALL Rs232c::Out6f(uint, uint data)
{
	//	0..75bps,	1..150bps,	2..300bps,	3..600bps,	4..1200bps
	//	5..2400bps,	6..4800bps,	7..9600bps,	8..19200bps
	int baudrate = (data & 0x0f);
	if (baudrate < 9)
	{
		if (config.baudrate != baudrate && hcom != INVALID_HANDLE_VALUE)
		{
			SetMode(baudrate);
		}
		config.baudrate = baudrate;
	}
}

uint IFCALL Rs232c::In20(uint)
{
	if (IsRs232cEnabled() && mode != clear)
	{
		//LOG0("In20\n");
		int count = (50 << (8 - config.baudrate)) + 50;
		scheduler->DelEvent(this);
		scheduler->AddEvent(count, this, STATIC_CAST(TimeFunc, &Rs232c::TimeEvent), 1);
		statuscount = 0;
	}
	return 0xff;
}

uint IFCALL Rs232c::In21(uint)
{
	uint status = 0xff;
	if (!checking && !setting) 
	{
		checking = true;
	
		// BIT 6	SYNC		BIT 5	FE
		// BIT 4	OE			BIT 3	PE
		// BIT 2	TxE			BIT 1	RxRDY
		if (IsRs232cEnabled())
		{
			//LOG0("In21\n");
			statuscount++;
			if (statuscount > 4)
			{
				// Xe[^X|[OAvdȂւ̑΍
				SetData(NULL, 0);
				statuscount = 0;
			}
			if (comstat.fDsrHold)	status &= ~DSR;
			if (comstat.fTxim)		status &= ~TXRDY;
		}
		checking = false;
	}
	return status;
}

uint IFCALL Rs232c::In6f(uint)
{
	return 0xf0 | config.baudrate;
}

// ---------------------------------------------------------------------------
//	RS-232C
//

bool Rs232c::Open(RS232C_CONFIG* pconfig)
{
	if (config.comport == pconfig->comport &&
		hcom != INVALID_HANDLE_VALUE)
	{
		if (config.baudrate != pconfig->baudrate)
		{
			SetMode(pconfig->baudrate);
		}
		config = *pconfig;
		return true;
	}

	Close();
	config = *pconfig;
	if (config.comport == 0)
		return true;

	TCHAR portname[16];
	wsprintf(portname, "COM%d", config.comport);
	hcom = CreateFile(portname, 
					  GENERIC_READ | GENERIC_WRITE, 
					  0, 
					  NULL, 
					  OPEN_EXISTING,
					  0, 
					  NULL);
	if (hcom == INVALID_HANDLE_VALUE)
	{
		LOG1("COM open error %08X\n", GetLastError());
		strcpy(message, "COM port open error");
		return false;
	}

	//if (!SetCommMask(hcom, EV_RXCHAR))
	//{
	//	Close();
	//	strcpy(message, "Comm mask setup error.");
	//	return false;
	//}

	memset(&dcb, 0, sizeof(DCB));
	dcb.DCBlength = sizeof(DCB);
	GetCommState(hcom, &dcb);

	if (!SetMode(config.baudrate))
	{
		Close();
		strcpy(message, "Comm state setup error.");
		return false;
	}

	if (!hthread)
	{
		running = true;
		hthread = (HANDLE)(_beginthreadex(NULL, 0, ThreadEntry, 
						   reinterpret_cast<void*> (this), 0, &idthread));
	}
	if (!hthread)
	{
		Close();
		strcpy(message, "Thread setup error.");
		return false;
	}
	
	SetupComm(hcom, 4096, 4096);
	PurgeComm(hcom, 
		PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR);

	return true;
}

void Rs232c::Close()
{
	if (hcom !=INVALID_HANDLE_VALUE)
	{
		PurgeComm(hcom, 
			PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR);
		CloseHandle(hcom);
	}

	//char buf[256];
	DWORD x = GetTickCount();

	if (hthread)
	{
		running = false;
		if (WAIT_TIMEOUT == WaitForSingleObject(hthread, 1000))
		{
			TerminateThread(hthread, 0);
			//sprintf(buf,"Terminate %d : ",GetTickCount() - x);
			//OutputDebugString(buf);
		}
		CloseHandle(hthread);
	}
	//sprintf(buf,"Close %08X : ", GetTickCount() - x);
	//OutputDebugString(buf);

	hcom = INVALID_HANDLE_VALUE;
	hthread = NULL;
	strcpy(message, "Disabled");
}

bool Rs232c::SetMode(int baudrate)
{
	PurgeComm(hcom, 
		PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR);

	char pa[16];
	char st[16];
	char result[64];
	dcb.fBinary			= TRUE;

	switch(parity)
	{
	case Parity::none:
		dcb.fParity	= FALSE;
		dcb.Parity	= NOPARITY;
		strcpy(pa, "NO");
		break;
	case Parity::even:
		dcb.fParity	= TRUE;
		dcb.Parity	= EVENPARITY;
		strcpy(pa, "EVEN");
		break;
	case Parity::odd:
		dcb.fParity	= TRUE;
		dcb.Parity	= ODDPARITY;
		strcpy(pa, "ODD");
		break;
	}
	switch(stop)
	{
	case 0:
	case 1:
		dcb.StopBits = ONESTOPBIT;
		strcpy(st, "1");
		break;
	case 2:
		dcb.StopBits = ONE5STOPBITS;
		strcpy(st, "1.5");
		break;
	case 3:
		dcb.StopBits = TWOSTOPBITS;
		strcpy(st, "2");
		break;
	}

	dcb.BaudRate	= (75 << (baudrate + 4)) / clock;
	dcb.ByteSize	= datalen;

	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;


	bool ret = (SetCommState(hcom, &dcb) != FALSE);
	if (ret)
	{
		strcpy(result, "Enabled");
	}
	else 
	{
		sprintf(result, "Comm state setup error 0x%08X.", GetLastError());
	}

	sprintf(message, "%s\nBaudrate : %d\nParity : %s\nDataSize : %d\nStopBit : %s\nBaudrate factor : x%d\n", 
		result, (75 << baudrate), pa, datalen, st, clock);
	LOG0(message);

	return ret;
}

void Rs232c::SetData(byte* data, uint count)
{
	EnterCriticalSection(&cs);

	for (uint i=0; i<count; i++)
	{
		databuf[(datapos + datacount) % DATABUF_SIZE] = data[i];
		datacount++;
		//LOG4("%c %02X,%d,%d\n", (isprint(data[i]) ? data[i] : '.'),  
		//	data[i], datapos, datacount);
	}

	uint in21 = access->In(0x21);
	if (datacount > 0)
	{
		DWORD t = timeGetTime();
		if ((in21 & 0x02) == 0)
		{
			//char bbb[256];
			//sprintf(bbb,"R:%02X %d %d\r\n",databuf[datapos], datapos, datacount);
			//OutputDebugString(bbb);
			//LOG4("\t%c %02X,%d,%d\n", (isprint(databuf[datapos]) ? databuf[datapos] : '.'),  
			//	databuf[datapos], datapos, datacount);
			access->Out(prequest, databuf[datapos]);
			datapos = (datapos + 1) % DATABUF_SIZE;
			datacount--;
			access->Out(prxrdy, 1);
			tick = t;
		}
		else if ((t - tick) > (DWORD)((130 >> config.baudrate) + 5))
		{
			access->Out(prxrdy, 1);
			tick = t;
		}
	}

	LeaveCriticalSection(&cs);
}

// ---------------------------------------------------------------------------
//	thread
//
uint Rs232c::ThreadMain()
{
	LOG0("Thread start\n");
	byte buf[DATABUF_SIZE];

	char bb[256];
	DWORD x = 0;
	DWORD x1 = 0;
	DWORD x2 = 0;
	DWORD x3 = 0;
	timeBeginPeriod(1);

	while (running)
	{
		DWORD	error = 0;
		DWORD	len = 0;
		if (ClearCommError(hcom, &error, &comstat))
		{
			x2 = timeGetTime();
			x3 = x2;
			//LOG2("\t\t%X, %d\n", error, stat.cbInQue);
			if (error == 0 && comstat.cbInQue > 0)
			{
				DWORD readlen = Min(DATABUF_SIZE, comstat.cbInQue);
				uint data = 0;
				if (ReadFile(hcom, buf, readlen, &len, NULL))
				{
					x3 = timeGetTime();
					//LOG2("\t\t-%d, %d\n", readlen, len);
					if (readlen != len)
					{
						LOG2("Data length error %d %d\n", readlen, len);
					}
				}
			}
			setting = true;
			SetData(buf, len);
			setting = false;
		}
		x1 = timeGetTime();
		Sleep(2);
		DWORD y = timeGetTime();
		sprintf(bb,"%d %d %d %d %d %d %d %d\r\n", x, y-x, y-x2, y-x3, y-x1, error, comstat.cbInQue, len);
		//OutputDebugString(bb);
		x = y;
	}
	LOG0("Thread end\n");
	timeEndPeriod(1);
	return 0;
}

//// ̏̓{c
//// CPUׂ悤ȍɂ悤Ǝv܂Ȃ
//uint Rs232c::ThreadMain()
//{
//	LOG0("Thread start\n");
//	byte buf[DATABUF_SIZE];
//
//	char bb[256];
//	DWORD x = 0;
//	DWORD x1 = 0;
//	DWORD x2 = 0;
//	DWORD x3 = 0;
//	timeBeginPeriod(1);
//	OVERLAPPED ov;
//	ZeroMemory(&ov, sizeof(OVERLAPPED));
//    ov.Offset = 0;
//    ov.OffsetHigh = 0;
//    ov.hEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
//
//	DWORD	event;
//	WaitCommEvent(hcom, &event, &ov);
//
//	while (running)
//	{
//		COMSTAT	stat;
//		DWORD	error = 0;
//		DWORD	len = 0;
//		ZeroMemory(&stat, sizeof(COMSTAT));
//		if (WaitForSingleObject(ov.hEvent, 1) == WAIT_OBJECT_0)
//		{
//		//if (WaitCommEvent(hcom, &event, &ov))
//		//{
//			ResetEvent(ov.hEvent);
//			//if (event & EV_RXCHAR) 
//			//{
//				if (ClearCommError(hcom, &error, &stat))
//				{
//					x2 = timeGetTime();
//					x3 = x2;
//					//LOG2("\t\t%X, %d\n", error, stat.cbInQue);
//					if (error == 0 && stat.cbInQue > 0)
//					{
//						DWORD readlen = Min(DATABUF_SIZE, stat.cbInQue);
//						uint data = 0;
//						if (ReadFile(hcom, buf, readlen, &len, NULL))
//						{
//							x3 = timeGetTime();
//							//LOG2("\t\t-%d, %d\n", readlen, len);
//							if (readlen != len)
//							{
//								LOG2("Data length error %d %d\n", readlen, len);
//							}
//						}
//					}
//					setting = true;
//					SetData(buf, len);
//					setting = false;
//				}
//			//}
//		}
//		x1 = timeGetTime();
//		Sleep(2);
//		DWORD y = timeGetTime();
//		sprintf(bb,"%d %d %d %d %d %d %d %d\r\n", x, y-x, y-x2, y-x3, y-x1, error, stat.cbInQue, len);
//		//OutputDebugString(bb);
//		x = y;
//	}
//	LOG0("Thread end\n");
//	timeEndPeriod(1);
//	return 0;
//}

uint __stdcall Rs232c::ThreadEntry(LPVOID arg)
{
	if (arg)
		return reinterpret_cast<Rs232c*> (arg)->ThreadMain();
	else
		return 0;
}

// ---------------------------------------------------------------------------
//	Timer event
//
void IOCALL Rs232c::TimeEvent(uint e)
{
	SetData(NULL, 0);
}

// ---------------------------------------------------------------------------
//	device description
//
const Device::Descriptor Rs232c::descriptor = { indef, outdef };

const Device::OutFuncPtr Rs232c::outdef[] = 
{
	STATIC_CAST(Device::OutFuncPtr, Reset),
	STATIC_CAST(Device::OutFuncPtr, Out20),
	STATIC_CAST(Device::OutFuncPtr, Out21),
	STATIC_CAST(Device::OutFuncPtr, Out30),
	STATIC_CAST(Device::OutFuncPtr, Out6f),
};

const Device::InFuncPtr Rs232c::indef[] = 
{
	STATIC_CAST(Device::InFuncPtr, In20),
	STATIC_CAST(Device::InFuncPtr, In21),
	STATIC_CAST(Device::InFuncPtr, In6f),
};
